#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Sep 30 14:15:17 2020

@author: tanya
"""

import os
import sys
from itertools import product
from collections import ChainMap
from sklearn.pipeline import Pipeline
from cdplib.pipeline_selector.PipelineSelector import PipelineSelector

sys.path.append(os.getcwd())


class GridSearchPipelineSelector(PipelineSelector):
    """
    """
    def __init__(self,
                 cost_func,
                 greater_is_better: bool,
                 trials_path: str,
                 backup_trials_freq: int = 1,
                 cross_val_averaging_func: callable = None):
        '''
        :param callable cost_func: function to minimize or maximize

        :param bool greater_is_better: when True
            cost_func is maximized, else minimized.

        :param str trials_path: path at which the trials object is saved
            in binary format. From the trials object we can
            select information about the obtained scores, score variations,
            and pipelines, and parameters tried out so far. If a trials object
            already exists at the given path, it is loaded and the
            search is continued, else, the search is started from
            the beginning.

        :param backup_trials_freq: frequecy in interations (trials)
            of saving the trials object at the trials_path.

        :param str log_path: Optional, when not provided logs to stdout.

        :param callable averaging_func: optional,
            when not provided set to mean. Function
            to aggregate the cross-validated values of the cost function.
            Classic situation is to take the mean,
            another example is, for example mean() - c*var().
        '''
        super().__init__(cost_func=cost_func,
                         greater_is_better=greater_is_better,
                         trials_path=trials_path,
                         backup_trials_freq=backup_trials_freq,
                         cross_val_averaging_func=cross_val_averaging_func)

    def run_trials(self):
        """
        """
        if not self._space_attached:
            err = "Parameter distribution space must be attached"
            self._logger.log_and_raise_error(err)

        self._trials = self._trials or []

        done_trial_ids = [{"name": trial["name"],
                           "param_set": trial["param_set"]}
                          for trial in self._trials]

        # list (generator) of (flattened) dictionaries
        # with all different combinations of
        # parameters for different pipelines
        # from the space definition.
        space_unfolded = ({"name": pipeline_dist["name"],
                           "pipeline": pipeline_dist["pipeline"],
                           "param_set": param_set}
                          for pipeline_dist in self._space
                          for param_set in
                          (dict(ChainMap(*tup)) for tup in
                           product(*[[{k: v} for v in
                                      pipeline_dist["params"][k]]
                                     for k in pipeline_dist["params"]])))

        for space_element in space_unfolded:

            trial_id = {"name": space_element["name"],
                        "param_set": space_element["param_set"]}

            if trial_id in done_trial_ids:
                continue

            result = self._objective(space_element)

            pipeline = space_element["pipeline"].set_params(
                    **space_element["param_set"])

            self._trials.append({"name": space_element["name"],
                                 "param_set": space_element["param_set"],
                                 "pipeline": pipeline,
                                 "result": result})

    @property
    def best_trial(self) -> dict:
        """
        """
        if self._trials is None:
            self._logger.error(("Trials object is empty. "
                                "Call run_trials method."))
        else:
            try:
                return max(gs._trials, key=lambda x: x["result"]["score"])
            except Exception as e:
                err = ("Could not retrieve the best trial. "
                       "Exit with error: {}".format(e))
                self._logger.log_and_raise_error(err)

    @property
    def best_trial_score(self) -> float:
        '''
        '''
        if self._trials is None:
            self._logger.error(("Trials object is empty. "
                                "Call run_trials method."))
        else:
            try:
                return self.best_trial["result"]["score"]
            except Exception as e:
                err = ("Could not retrieve the best trial. "
                       "Exit with error: {}".format(e))
                self._logger.log_and_raise_error(err)

    @property
    def best_trial_score_variance(self) -> float:
        '''
        '''
        if self._trials is None:
            self._logger.error(("Trials object is empty. "
                                "Call run_trials method."))
        else:
            try:
                return self.best_trial["result"]["score_variance"]
            except Exception as e:
                err = ("Could not retrieve the best trial. "
                       "Exit with error: {}".format(e))
                self._logger.log_and_raise_error(err)

    @property
    def best_trial_pipeline(self) -> Pipeline:
        '''
        '''
        if self._trials is None:
            self._logger.error(("Trials object is empty. "
                                "Call run_trials method."))
        else:
            try:
                return self.best_trial["pipeline"]
            except Exception as e:
                err = ("Could not retrieve the best trial. "
                       "Exit with error: {}".format(e))
                self._logger.log_and_raise_error(err)


if __name__ == "__main__":

    # Small test

    from sklearn.datasets import load_breast_cancer
    from sklearn.metrics import accuracy_score
    from cdplib.gridsearch.space_sample import space
    from cdplib.log import Log

    logger = Log("GridSearchPipelineSelector__TEST:")

    logger.info("Start test")

    data_loader = load_breast_cancer()

    X = data_loader["data"]
    y = data_loader["target"]

    gs = GridSearchPipelineSelector(cost_func=accuracy_score,
                                    greater_is_better=True,
                                    trials_path="trials_TEST.pkl")

    gs.attach_space(space=space)
    gs.attach_data(X_train=X, y_train=y, cv=5)

    gs.run_trials()

    logger.info("Best trial: {}".format(gs.best_trial))
    logger.info("Best trial: {}".format(gs.best_trial_pipeline))

    logger.info("End test")
